﻿using System;
using System.Text;
using Sandbox.ModAPI;
using VRage.Game.Components;
using VRage.ModAPI;
using Sandbox.Common.ObjectBuilders;
using VRage.ObjectBuilders;
using Sandbox.Definitions;
using VRage.Game;
using Sandbox.Game.EntityComponents;
using VRage.Game.Entity;
using Sandbox.Game;
using VRageMath;
using SpaceEngineers.Game.ModAPI;
using VRage.Game.ModAPI;
using System.Collections.Generic;
using Sandbox.Game.Entities;
using Sandbox.ModAPI.Interfaces;
using System.Linq;
using Ingame = Sandbox.ModAPI.Ingame;
using Sandbox.Game.AI;

namespace Cheetah.AI {

    public sealed class FighterBot : BotBase {

        //Таймер обновления AI, используем минимальное и максимальное значене чтобы размазать вызовы и вслучае большого количества дронов не вызывать фризы
        private int AI_Update_Timmer = 120;
        //Дефолтовое значение расчитываем по формуле базовый UPS * количество секунд задержки т.е. 60*5 это примерно каждые 5 секунд
        private int AI_Update_Timmer_Min = 60*1;
        private int AI_Update_Timmer_Max = 60*2;
        //Дистанция на которую перемножаем вектор направления для получения удаленной точки оступления
        private float AI_Fly_Distance = 100000;
        //Кеновский AI тупит если находится слишком далеко, поэтому ограничиваем радиус его активации и подменяем его автопилотов
        public float AI_Maximum_Activation_Range = 2000;
        //Рандомизатор используемый скилючительно для AI
        Random AI_Randomizer = new Random();
        //Переменная отображающая прогрузился ли кастомный AI
        public bool AI_Initialize = false;
        //Флаг определяющий что корабль получил слишком большие повреждения и нужно сваливать
        public bool Too_Much_Damage = false;
        //Флаг определяющий что корабль бежит с поля боя
        public bool Need_Escape = false;
        //Безумные ребята никогда не отступают =)
        public bool AI_is_Crazy = false;

        static public readonly BotTypes BotType = BotTypes.Fighter;
        private bool Damaged = true;
        public bool KeenAILoaded { get; private set; }

        FighterSettings FighterSetup;
        struct FighterSettings {

            public string Preset;
            public bool CallHelpOnDamage;
            public bool AssignToPirates;
            public bool DelayedAIEnable;
            public bool AmbushMode;
            public float MaxSpeed;
            public float SeekDistance;
            public float AIActivationDistance;
            public int PlayerPriority;
            public int CallHelpProbability;
            //Максимальный шанс вызова подкрепления, в процентах
            public static int CallHelpProbabilityMax;
            //Процент пределяющий побежит корабль или нет
            public int Run_Chance;

            /// <summary>
            /// Fills out the empty values in struct with default values, and leaves filled values untouched.
            /// </summary>
            public void Default(bool RandomizeCallHelp = true) {
                if (Preset == default(string)) { Preset = "DefaultDirect"; }
                if (AssignToPirates == default(bool)) { }
                if (AmbushMode == default(bool)) { }
                if (DelayedAIEnable == default(bool)) { }
                if (SeekDistance == default(float)) { SeekDistance = 10000; }
                if (PlayerPriority == default(int)) { PlayerPriority = 10; }
                if (CallHelpProbability == default(int)) { CallHelpProbability = CallHelpProbabilityMax; }
                if (CallHelpOnDamage == default(bool) || RandomizeCallHelp) { this.RandomizeCallHelp(); }
                if (Run_Chance == default(int)) { Run_Chance = 75; }
            }

            private void RandomizeCallHelp() {
                Random Rand = new Random();
                int random = Rand.Next(0, 101);

                if (random <= CallHelpProbability) { CallHelpOnDamage = true; }
            }

            public override string ToString() {
                return $"Preset='{Preset}|CallHelp={CallHelpOnDamage}|AssignToPirates={AssignToPirates}|SeekDistance={SeekDistance}|PlayerPriority={PlayerPriority}";
            }
        }

        public FighterBot(IMyCubeGrid Grid) : base(Grid) { }

        public override bool Init(IMyRemoteControl RC = null) {

            if(!base.Init(RC)) return false;
            Update |= MyEntityUpdateEnum.EACH_10TH_FRAME | MyEntityUpdateEnum.EACH_100TH_FRAME;

            if (FighterSetup.CallHelpOnDamage) OnDamaged += DamageHandler;

            RC.Name = DroneNameProvider;
            MyAPIGateway.Entities.SetEntityName(RC as IMyEntity, true);

            AI_is_Crazy = CheckRammingBehavior();

            return true;
        }
        
        //Проверяем может ли боевой Ai таранить если кончилось оружие... возможно есть более правильный способ проверки но толковой документации к инжинерам нихера нет посему сделал так... не надежно зато просто
        public bool CheckRammingBehavior() {
            if (FighterSetup.Preset == "EEMDefaultSmall" || FighterSetup.Preset == "EEMGrinder" || FighterSetup.Preset == "EEMHomingMine") { 
                return true;
            }
            else { return false; }
        }

        private System.Diagnostics.Stopwatch _sw = new System.Diagnostics.Stopwatch();
        //Основная функция скрипта, вызывается каждый апдэйт т.е. 60 раз в секунду(при хорошей скорости симуляции)
        public override void Main() {
            try {

                //Поскольку 60 раз в секунду обновлять параметры "мозгов" нам нахрен не нужно, используем простенький таймер и если время пришло...
                if (AI_Update_Timmer <= 0) {

                    _sw.Reset();
                    _sw.Start();

                    if (!AI_Initialize) {

                        //Выставляем максимальную скорость для корабля в зависимости от максимальной скорости в мире и ограничиваем ее настройками корабля на случай если у каких-то наркоманов максимальная скорость 99999 -_-
                        FighterSetup.MaxSpeed = RC.GetSpeedCap() * 0.98f;
                        if (FighterSetup.MaxSpeed > 500) { FighterSetup.MaxSpeed = 500; }

                        //Считаем шанс побега при получении критических повреждений, для пресетов которые пытаются таранить если нет оружия делаем шанс побега минимальным, это у нас дроны либо совсем отмороженные ребята =)
                        if (AI_is_Crazy) { FighterSetup.Run_Chance = AI_Randomizer.Next(0, 15); }
                        else { FighterSetup.Run_Chance = AI_Randomizer.Next(25, 75); }
                        
                        Digi.Log.Info(String.Format("{0} - FighterAI_Initialized - [AI_Update: {1} - {2} sec], [AgressionDistance: {3} km], [MaxSpeed: {4} m/s], [CallHelpChance: {5}%], [CallHelpOnDamage: {6}], [DamageRunChance: {7}]" , 
                            GridName,
                            Math.Round((float)AI_Update_Timmer_Min/60, 2),
                            Math.Round((float)AI_Update_Timmer_Max/60, 2),
                            Math.Round(FighterSetup.SeekDistance/1000, 2),
                            FighterSetup.MaxSpeed,
                            FighterSetup.CallHelpProbability,
                            FighterSetup.CallHelpOnDamage.ToString(),
                            FighterSetup.Run_Chance
                        ));
                        
                        AI_Initialize = true;
                    }

                    if (AISessionCore.AI_Debug) { Digi.Log.Info(Environment.NewLine + GridName + " - FighterAI_Update - Start"); }
				    
                    //...вызываем обработку искуственного интеллекта...
                    FighterAI_Update();

                    //...рандомизируем время следующего вызова дабы размазать возможную нагрузку и перезапускаем таймер.
                    int Random_Time = AI_Randomizer.Next(AI_Update_Timmer_Min, AI_Update_Timmer_Max);
                    AI_Update_Timmer = Random_Time;

                    _sw.Stop();
                    if (AISessionCore.AI_Debug) { 
                        Digi.Log.Info(GridName + " - FighterAI_Update - End, Next: " + Random_Time + ", ProcessingTime: " + _sw.Elapsed + Environment.NewLine); 
                    }
                
                }
                else { AI_Update_Timmer--; }

            }
            catch (Exception ex) {
                Digi.Log.Error(ex.Message +  ex.StackTrace);
            }
            
        }

        //Основная функция управления интелектом дрона
        private void FighterAI_Update() {

			try {

                //Получаем текущий статус сетки...
                GridStatusData GSD = GetStatusData();

                if (AISessionCore.AI_Debug) {
                    string weapon_text = "";
                    string reactors_text = "";
                    string gyro_text = "";
                    string rc_text = "";
                    string total_hp_text = "";

                    if (GSD.Weapon_Left == 0 && !AI_is_Crazy) { weapon_text = "0 - Need retreat"; }
                    else if (GSD.Weapon_Left == 0) { weapon_text = "0 - Ramming mod: ON"; }
                    else { weapon_text = GSD.Weapon_Left + " - OK"; }

                    if (GSD.EnergyBlock_Left < 1f) { reactors_text = "One with " + Math.Round(GSD.EnergyBlock_Left*100, 2) + "% hp, need retreat"; }
                    else { reactors_text = GSD.EnergyBlock_Left + " - OK"; }

                    if (GSD.Gyro_Left < 1f) { gyro_text = "One with " + Math.Round(GSD.Gyro_Left*100, 2) + "% hp, need retreat"; }
                    else { gyro_text = GSD.Gyro_Left + " - OK"; }

                    if (GSD.RC_HP_left <= 0.6f) { rc_text = Math.Round(GSD.RC_HP_left*100, 2) + " < 50% - Need retreat"; }
                    else { rc_text = Math.Round(GSD.RC_HP_left*100, 2) + "% - OK"; }

                    if (GSD.Total_Health_Left <= 0.6f) { total_hp_text = Math.Round(GSD.Total_Health_Left*100, 2) + " < 50% - Need retreat"; }
                    else { total_hp_text = Math.Round(GSD.Total_Health_Left*100, 2) + "% - OK"; }

                    Digi.Log.Info(String.Format("{0} - GridStatusData: [Preset: {1}], [Weapons: {2}], [EnergyBlocks: {3}], [Gyro: {4}], [RC health: {5}], [Total health: {6} - ({7}/{8})], [Escape: {9}]", 
                        GridName,
                        FighterSetup.Preset,
                        weapon_text,
                        reactors_text,
                        gyro_text,
                        rc_text,
                        total_hp_text,
                        GSD.Total_Health,
                        GSD.Total_Blocks,
                        Need_Escape.ToString()
                    )); 

                }

                //...если еще не получили критические повреждения...
                if (!Too_Much_Damage) {

                    //...если не осталось рабочего оружия и AI не отморозок то со 100%-ным шансом бежим...
                    if (GSD.Weapon_Left == 0 && !AI_is_Crazy) { 
                        Too_Much_Damage = true; 
                        Need_Escape = true;
                        Digi.Log.Info(GridName + " - No weapon left, need escape...");
                    }
                    //...проверяем получил ли корабль достаточно повреждений...
                    else if (GSD.EnergyBlock_Left < 1f
                        || GSD.Gyro_Left < 1f 
                        || GSD.RC_HP_left < 0.6f
                        || GSD.Total_Health_Left < 0.5f) { 

                        Too_Much_Damage = true; 
                    }

                    //...если теперь он критический то однократно рассчитываем побежит бот или нет...
                    if (Too_Much_Damage && !Need_Escape) {
                        int run_chance = AI_Randomizer.Next(0, 101);
                        if (run_chance <= FighterSetup.Run_Chance) { 
                            Need_Escape = true; 
                            Digi.Log.Info(GridName + " - RunChance: "+ run_chance + " <= " + FighterSetup.Run_Chance + " - Too much damage... Escape... I NEED ESCAPE!!!");
                        }
                        else { Digi.Log.Info(GridName + " - RunChance: "+ run_chance + " > " + FighterSetup.Run_Chance + " - Damage is nothing for me, i will continue fight!"); }

                    }
                }

                //Если не нужно убегать то...
                if (!Need_Escape) {

                    //...ищем враждебные объекты в заданном радиусе...
                    List<CustomTargetData> RadarData = LookForEnemies(FighterSetup.SeekDistance);

                    //RadarDetectedEntity.Type = Ingame.MyDetectedEntityType.
                    if (AISessionCore.AI_Debug) { Digi.Log.Info(GridName + " - TargetsCount: " + RadarData.Count + ", Range: " + FighterSetup.SeekDistance); }

                    //...если что-то нашли...
                    if (RadarData.Count > 0) {

                        //...получаем ближайший враждебный объект с высшим приоритетом...
                        CustomTargetData ClosestTargetData = RadarData.OrderBy(x => x._Priority).ThenBy(x => x._Distance).FirstOrDefault();
                    
                        //...если это ни ничто...
                        if (!ClosestTargetData._EntityInfo.IsEmpty()) {

                            //...получаем игровой объект на основе id ближайшего враждебного объекта...
                            IMyEntity EnemyEntity = ClosestTargetData._Entity as IMyEntity;

                            //...если удалось получить объект...
                            if (EnemyEntity != null) {

                                //[КОСТЫЛЬ]
                                //Поскольку кеновкий AI отказывается лететь к игроку на расстоянии более 4-5км, мы вручную заставляем эту скотину прилететь куда нужно... -_-

                                //...определяем дистанцию до ближайшего вражеского объекта и если она слишком высока...
                                double DistanceToEnemy = GridPosition.DistanceTo(ClosestTargetData._EntityInfo.Position);
                                if (DistanceToEnemy > AI_Maximum_Activation_Range) {

                                    //...отправляем дрона на сближение.
                                    AI_FlyToPoint(ClosestTargetData._EntityInfo.Position, FighterSetup.MaxSpeed, "Combat_AI_Target");

                                    if (AISessionCore.AI_Debug) { 
                                        Digi.Log.Info(GridName + " - Combat_AI_FlyToTarget: " + ClosestTargetData._Name + ", Priority: " + ClosestTargetData._Priority + ", Distance: " + ClosestTargetData._Distance); 
                                    }
                                }
                                //...иначе активируем AI...
                                else {

                                    //...задаем параметры и активируем AI, он перехватит управление автопилотом и начнет охоту за игроком.
                                    (RC as MyRemoteControl).SetAutoPilotSpeedLimit(FighterSetup.MaxSpeed);
				                    //MyVisualScriptLogicProvider.SetDroneBehaviourFull(RC.Name, presetName: FighterSetup.Preset, activate: true, maxPlayerDistance: FighterSetup.SeekDistance, playerPriority: FighterSetup.PlayerPriority, assignToPirates: FighterSetup.AssignToPirates);
                                    MyVisualScriptLogicProvider.SetDroneBehaviourFull(RC.Name, presetName: FighterSetup.Preset, activate: true, maxPlayerDistance: FighterSetup.SeekDistance, playerPriority: 0, assignToPirates: FighterSetup.AssignToPirates);
                                    if (FighterSetup.AmbushMode == true) MyVisualScriptLogicProvider.DroneSetAmbushMode(RC.Name, ambushModeOn: true);
                                    MyVisualScriptLogicProvider.DroneSetTarget(RC.Name, ClosestTargetData._Entity);
                                
                                    //...метим что наш безмозглый мозг активирован.
                                    KeenAILoaded = true;

                                    if (AISessionCore.AI_Debug) { Digi.Log.Info(GridName + " - Combat_AI_Enabled, Target: " + ClosestTargetData._Name + ", Priority: " + ClosestTargetData._Priority); }
                            
                                }

                            }
                        }
 
                    }
                    //Иначе отступаем...
                    else { 

                        //...получаем вектор направления блока удаленного управления, перемножаем на дистанцию получая тем самым удаленную точку
                        Vector3D RetreatPoint = RC.WorldMatrix.Forward * AI_Fly_Distance; 
                    
                        //...пызываем функцию полета к сладосному диспавну
                        AI_FlyToPoint(RetreatPoint, RC.GetSpeedCap(), "Retreat");

                        if (AISessionCore.AI_Debug) { Digi.Log.Info(GridName + " - Combat_AI_Disabled - Retrea."); }
                
                    }
                }
                //Иначе бежим...
                else { 

                    //...получаем вектор направления блока удаленного управления, перемножаем на дистанцию получая тем самым удаленную точку
                    Vector3D RetreatPoint = RC.WorldMatrix.Forward * AI_Fly_Distance; 
                    
                    //...пызываем функцию полета к сладосному диспавну
                    AI_FlyToPoint(RetreatPoint, RC.GetSpeedCap(), "Escape");

                    if (AISessionCore.AI_Debug) { Digi.Log.Info(GridName + " - Combat_AI_Disabled - Escape"); }
                
                }

			}
			catch (Exception ex) {
				Digi.Log.Error(ex.Message + Environment.NewLine + ex.StackTrace);
			}
        }
		
        //Функция полета к заданной точке
        public void AI_FlyToPoint(Vector3D TargetPoint, float MaxSpeed, string TargetPointName = "FlyTargetPoint") {
            
            //Отрубаем киновский AI чтобы не мешал автопилоту
            MyVisualScriptLogicProvider.DroneTargetClear(RC.Name);
            MyVisualScriptLogicProvider.DroneWaypointClear(RC.Name);
            MyVisualScriptLogicProvider.SetDroneBehaviourFull(RC.Name, presetName: FighterSetup.Preset, activate: false, maxPlayerDistance: 0, playerPriority: 0, assignToPirates: FighterSetup.AssignToPirates);
            if (FighterSetup.AmbushMode == true) MyVisualScriptLogicProvider.DroneSetAmbushMode(RC.Name, ambushModeOn: true);
            //MyVisualScriptLogicProvider.DroneSetTarget(RC.Name, Target); //Не работает, либо я не разобрался как правильно пользоваться
            //MyVisualScriptLogicProvider.DroneSetRetreatPosition(RC.Name, RetreatPoint); //Не работает, либо я не разобрался как правильно пользоваться

            //На всякий случай выставляем значение переменной, тут она конечно уже не используется но мало ли, вдруг кому потребуется получить состояние AI
            KeenAILoaded = false;

            //Очищаем списой вэйпоинтов автопилота и добавляем нужный
            RC.ClearWaypoints();
		    RC.AddWaypoint(TargetPoint, TargetPointName);

            //Задаем параметры полета и активируем автопилот
		    (RC as MyRemoteControl).ChangeFlightMode(MyRemoteControl.FlightMode.OneWay);
		    (RC as MyRemoteControl).SetAutoPilotSpeedLimit(MaxSpeed);
            (RC as MyRemoteControl).SetCollisionAvoidance(true);
		    RC.SetAutoPilotEnabled(true);
        }

        private void DamageHandler(IMySlimBlock Block, MyDamageInformation Damage) {

            IMyPlayer Damager;
            ReactOnDamage(Block, Damage, CalmdownTime, out Damager);

            if (Damaged) { return; }
            Damaged = true;

            foreach (IMyTimerBlock Timer in Term.GetBlocksOfType<IMyTimerBlock>(collect: x => x.IsFunctional && x.Enabled
            && x.CustomName.Contains("Damage"))) {
                Timer.Trigger();
            }

            if (!FighterSetup.CallHelpOnDamage) { return; }

            foreach (IMyTimerBlock Timer in Term.GetBlocksOfType<IMyTimerBlock>(collect: x => x.IsFunctional && x.Enabled
            && x.CustomName.Contains("Security"))) {
                Timer.Trigger();
            }
        }



        protected override bool ParseSetup()
        {
            if (BotBase.ReadBotType(RC) != BotType) return false;

            List<string> CustomData = RC.CustomData.Trim().Replace("\r\n", "\n").Split('\n').ToList();
            foreach (string DataLine in CustomData)
            {
                if (DataLine.Contains("EEM_AI")) continue;
                if (DataLine.Contains("Type")) continue;
                var Data = DataLine.Trim().Split(':');
                Data[0] = Data[0].Trim();
                Data[1] = Data[1].Trim();

                switch (Data[0])
                {
                    case "Faction":
                        break;
                    case "Preset":
                        FighterSetup.Preset = Data[1];
                        break;
                    case "DelayedAI":
                        if (!bool.TryParse(Data[1], out FighterSetup.DelayedAIEnable))
                        {
                            DebugWrite("ParseSetup", "AI setup error: DelayedAI cannot be parsed");
                            return false;
                        }
                        break;
                    case "AssignToPirates":
                        if (!bool.TryParse(Data[1], out FighterSetup.AssignToPirates))
                        {
                            DebugWrite("ParseSetup", "AI setup error: AssignToPirates cannot be parsed");
                            return false;
                        }
                        break;
                    case "AmbushMode":
                        if (!bool.TryParse(Data[1], out FighterSetup.AmbushMode))
                        {
                            DebugWrite("ParseSetup", "AI setup error: AmbushMode cannot be parsed");
                            return false;
                        }
                        break;
                    case "SeekDistance":
                        if (!float.TryParse(Data[1], out FighterSetup.SeekDistance))
                        {
                            DebugWrite("ParseSetup", "AI setup error: SeekDistance cannot be parsed");
                            return false;
                        }
                        break;
                    case "ActivationDistance":
                        if (!float.TryParse(Data[1], out FighterSetup.AIActivationDistance))
                        {
                            DebugWrite("ParseSetup", "AI setup error: ActivationDistance cannot be parsed");
                            return false;
                        }
                        break;
                    case "PlayerPriority":
                        if (!int.TryParse(Data[1], out FighterSetup.PlayerPriority))
                        {
                            DebugWrite("ParseSetup", "AI setup error: PlayerPriority cannot be parsed");
                            return false;
                        }
                        break;
                    case "CallHelpProbability":
                        int probability;
                        if (!int.TryParse(Data[1], out probability))
                        {
                            DebugWrite("ParseSetup", "AI setup error: CallHelpProbability cannot be parsed");
                            return false;
                        }
                        else if (probability < 0 || probability > 100)
                        {
                            DebugWrite("ParseSetup", "AI setup error: CallHelpProbability out of bounds. Must be between 0 and 100");
                            return false;
                        }
                        else {
                            //Ограничиваем шанс вызова подмоги воизбежание бесконечного спавна
                            if (probability > FighterSettings.CallHelpProbabilityMax) { FighterSetup.CallHelpProbability = FighterSettings.CallHelpProbabilityMax; }
                            else { FighterSetup.CallHelpProbability = probability; }
                        }
                        break;
                    case "ThrustMultiplier":
                        float multiplier;
                        if (!float.TryParse(Data[1], out multiplier))
                        {
                            DebugWrite("ParseSetup", "AI setup error: ThrustMultiplier cannot be parsed");
                            return false;
                        }
                        else
                        {
                            ApplyThrustMultiplier(multiplier);
                        }
                        break;
                    default:
                        DebugWrite("ParseSetup", $"AI setup error: Cannot parse '{DataLine}'");
                        return false;
                }
            }
            FighterSetup.Default();
            return true;
        }

        protected override void DebugWrite(string Source, string Message, string DebugPrefix = "FighterBot.")
        {
            base.DebugWrite(Source, Message, DebugPrefix);
        }
    }
}