﻿using System;
using System.Text;
using Sandbox.ModAPI;
using VRage.Game.Components;
using VRage.ModAPI;
using Sandbox.Common.ObjectBuilders;
using VRage.ObjectBuilders;
using Sandbox.Definitions;
using VRage.Game;
using Sandbox.Game.EntityComponents;
using VRage.Game.Entity;
using Sandbox.Game;
using VRageMath;
using SpaceEngineers.Game.ModAPI;
using VRage.Game.ModAPI;
using System.Collections.Generic;
using Sandbox.Game.Entities;
using Sandbox.ModAPI.Interfaces;
using System.Linq;
using Ingame = Sandbox.ModAPI.Ingame;


namespace Cheetah.AI {
    public sealed class FreighterBot : BotBase {

        //Таймер обновления AI, используем минимальное и максимальное значене чтобы размазать вызовы и вслучае большого количества дронов не вызывать фризы
        private int AI_Update_Timmer = 60;
        //Дефолтовое значение расчитываем по формуле базовый UPS * количество секунд задержки т.е. 60*5 это примерно каждые 5 секунд
        private int AI_Update_Timmer_Min = 60*10;
        private int AI_Update_Timmer_Max = 60*30;
        //Дистанция на которую перемножаем вектор направления для получения удаленной точки оступления
        private float AI_Fly_Distance = 100000;
        //Рандомизатор используемый скилючительно для AI
        private Random AI_Randomizer = new Random();
        //Переменная отображающая прогрузился ли кастомный AI
        public bool AI_Initialize = false;

        //Задаем минимальную и максимальную скорости в режиме покоя и режиме бегства, перезаписывается в функции Main() при инициализации
        public float AI_DeepSpaceSpeed_Min = 5.0f;
        public float AI_DeepSpaceSpeed_Max = 25.0f;
        public float AI_RunSpeed_Min = 75.0f;
        public float AI_RunSpeed_Max = 95.0f;
        
        //Справочник для хранения связки игрок - оставшееся время
        public Dictionary<IMyPlayer,int> AI_TooClose_TimmerList = new Dictionary<IMyPlayer,int>();
        //Время таймера, высчитывается по форуле UPS * количество секунд, т.е. 60*80 это примерно 80 секунд.
        public int AI_TooClose_TimmerDefault_Min = 60*80;
        public int AI_TooClose_TimmerDefault_Max = 60*150;

        static public readonly BotTypes BotType = BotTypes.Freighter;
        FreighterSettings FreighterSetup;
        struct FreighterSettings {
            public bool FleeOnlyWhenDamaged;
            public float FleeTriggerDistance;
            public float FleeSpeedRatio;
            public float FleeSpeedCap;

            public void Default() {
                if (FleeOnlyWhenDamaged == default(bool)) FleeOnlyWhenDamaged = false;
                if (FleeTriggerDistance == default(float)) FleeTriggerDistance = 1000;
                if (FleeSpeedRatio == default(float)) FleeSpeedRatio = 1.0f;
                if (FleeSpeedCap == default(float)) FleeSpeedCap = 300;
            }
        }
        private bool IsFleeing = false;
        private bool FleeTimersTriggered = false;

        public FreighterBot(IMyCubeGrid Grid) : base(Grid) { }

        public override bool Init(IMyRemoteControl RC = null) {

            if (!base.Init(RC)) return false;
            Update |= MyEntityUpdateEnum.EACH_100TH_FRAME;
            OnDamaged += DamageHandler;
            OnBlockPlaced += FreighterBlockPlacedHandler;
            Grid.OnBlockRemoved += FreighterBlockRemoveHandler;

            return true;
        }

        //Небольшой костыль, сбрасываем все таймеры до объявления войны если кто-то посмел удалить блок... я не знаю как получить того кто удалил блок и мне влом разбираться т.к. документация к API инжинеров крааайне скудная.
        public void FreighterBlockRemoveHandler(IMySlimBlock Block) {
            if (Block == null) return;

            try {

                if (AISessionCore.AI_Debug) { Digi.Log.Info(GridName + " - BlockRemoved - Reset all war timmers"); }
                
                List<IMyPlayer> AITCTL_keys = new List<IMyPlayer>(AI_TooClose_TimmerList.Keys);
                for (int i = 0; i < AITCTL_keys.Count; i++) { AI_TooClose_TimmerList[AITCTL_keys[i]] = 0; }

            }
            catch (Exception Scrap) { Grid.LogError("BlokPlaedHandler", Scrap); }
        }

        public void FreighterBlockPlacedHandler(IMySlimBlock Block)
        {
            if (Block == null) return;

            try
            {
                IMyPlayer Builder = null;
                IMyFaction Faction = null;
                if (Block.IsPlayerBlock(out Builder))
                {
                    Faction = Builder.GetFaction();
                    if (Faction != null)
                    {
                        RegisterHostileAction(Faction, CalmdownTime);
                        if (!FleeTimersTriggered) { TriggerFleeTimers(); }
                    }
                }
            }
            catch (Exception Scrap)
            {
                Grid.LogError("BlokPlaedHandler", Scrap);
            }
        }

        private void DamageHandler(IMySlimBlock Block, MyDamageInformation Damage) {
            try {

                IMyPlayer Damager;
                ReactOnDamage(Block, Damage, CalmdownTime, out Damager);

                //Если защитные системы еще не активированы и удалось определить кто нанес урон, активируем защитные системы
                if (!FleeTimersTriggered && Damager != null) { TriggerFleeTimers(); }
            }
            catch (Exception Scrap) {
                Grid.LogError("DamageHandler", Scrap);
            }
        }

        private System.Diagnostics.Stopwatch _sw = new System.Diagnostics.Stopwatch();
        //Основная функция скрипта, вызывается каждый апдэйт т.е. 60 раз в секунду(при хорошей скорости симуляции)
        public override void Main() {
            try {
                //Поскольку 60 раз в секунду обновлять параметры "мозгов" нам нахрен не нужно, используем простенький таймер и если время пришло...
                if (AI_Update_Timmer <= 0) {

                    _sw.Reset();
                    _sw.Start();

                    if (!AI_Initialize) {

                        //Выставляем максимальную скорость для корабля в зависимости от максимальной скорости в мире и ограничиваем ее настройками корабля на случай если у каких-то наркоманов максимальная скорость 99999 -_-
                        float World_Max_Speed = RC.GetSpeedCap();
                        if (World_Max_Speed > FreighterSetup.FleeSpeedCap) { World_Max_Speed = FreighterSetup.FleeSpeedCap; }

                        AI_DeepSpaceSpeed_Min = World_Max_Speed * 0.25f;
                        AI_DeepSpaceSpeed_Max = World_Max_Speed * 0.50f;
                        AI_RunSpeed_Min = World_Max_Speed * 0.50f;
                        AI_RunSpeed_Max = World_Max_Speed * 0.75f;

                        Digi.Log.Info(String.Format("{0} - FreighterAI_Initialized - [AI_Update: {1} - {2} sec], [DeepSpaceSpeed: {3} - {4} m/s], [RunSpeed: {5} - {6} m/s], [AgressionTimmer: {7} - {8} sec], [AgressionDistance: {9} km]" , 
                            GridName,
                            Math.Round((float)AI_Update_Timmer_Min/60, 2),
                            Math.Round((float)AI_Update_Timmer_Max/60, 2),
                            AI_DeepSpaceSpeed_Min,
                            AI_DeepSpaceSpeed_Max,
                            AI_RunSpeed_Min,
                            AI_RunSpeed_Max,
                            Math.Round((float)AI_TooClose_TimmerDefault_Min/60, 2),
                            Math.Round((float)AI_TooClose_TimmerDefault_Max/60, 2),
                            Math.Round(FreighterSetup.FleeTriggerDistance/1000, 2)
                        ));

                        AI_Initialize = true;
                    }

                    if (AISessionCore.AI_Debug) { Digi.Log.Info(Environment.NewLine + GridName + " - FreighterAI_Update - Start"); }
                    
                    //...вызываем обработку искуственного интеллекта...
                    FreighterAI_Update();

                    //...рандомизируем время следующего вызова дабы размазать возможную нагрузку и перезапускаем таймер...
                    int Random_Time = AI_Randomizer.Next(AI_Update_Timmer_Min, AI_Update_Timmer_Max);
                    AI_Update_Timmer = Random_Time;

                    //...обновляем таймеры игроков которые находятся слишком близко.
                    List<IMyPlayer> AITCTL_keys = new List<IMyPlayer>(AI_TooClose_TimmerList.Keys);
                    for (int i = 0; i < AITCTL_keys.Count; i++) { AI_TooClose_TimmerList[AITCTL_keys[i]] -= AI_Update_Timmer; }
                    
                    _sw.Stop();
                    if (AISessionCore.AI_Debug) { 
                        Digi.Log.Info(GridName + " - FreighterAI_Update - End, Next: " + Random_Time + ", ProcessingTime: " + _sw.Elapsed + Environment.NewLine); 
                    }
                
                }
                else { AI_Update_Timmer--; }

            }
            catch (Exception ex) { Digi.Log.Error(ex.Message + ex.StackTrace); }
            
        }


        //Основная функция управления интелектом дрона
        private void FreighterAI_Update() {

			try {

                //[Обрабатываем объявление войны если игрок слишком долго находится близко]
                //Данный способ обработки исключает слепую зону радара и у игрока не выйдет, сидя внутри корабля пересоздать фракцию для заключения мира а затем начать его захват, так же благодаря слепой зоне радара корабль сбросит скорость что позволит дронам вызванным на подмогу быстрее окружить его и "выкрить" игрока.
                //Получаем список всех игроков и перебираем его...
                List<IMyPlayer> PlayersOnlineList = new List<IMyPlayer>();
                MyAPIGateway.Players.GetPlayers(PlayersOnlineList);
                foreach (IMyPlayer tmp_Player in PlayersOnlineList) {

                    //...получаем контролируемый игроком объект и его позицию и дистанцию до него...
                    IMyEntity CP_Entity = tmp_Player.Controller.ControlledEntity.Entity;
                    Vector3D CP_Entity_Pos = CP_Entity.GetPosition();
                    double CP_Entity_Distance = GridPosition.DistanceTo(CP_Entity_Pos);
                    
                    //...получаем фракцию игрока, если она есть...
                    IMyFaction tmp_Faction = tmp_Player.GetFaction();
                    if (tmp_Faction != null) {

                        //...если дистанция до контролируемого игроком объекта слишком мала...
                        if (CP_Entity_Distance <= FreighterSetup.FleeTriggerDistance) {

                            //...по умолчанию считаем что объявлять войну не нужно...
                            bool Need_War = false;

                            //...если игрок и так враждебен то сразу метим что надо объявить войну...
                            if (OwnerFaction != null && OwnerFaction.IsHostileTo(tmp_Faction)) { Need_War = true; }
                            //...иначе обрабатываем таймеры...
                            else {
                                //...если игрок уже есть в писке...
                                if (AI_TooClose_TimmerList.ContainsKey(tmp_Player)) {

                                    //...проверяем оставшееся время и если оно вышло или игрок уже враждебен...
                                    int tmp_AITCT = AI_TooClose_TimmerList[tmp_Player];
                                    if (tmp_AITCT <= 0) { 
                                        //...метим что нужно объявить войну...
                                        Need_War = true; 
                                        //...удаляем отработавший таймер.
                                        AI_TooClose_TimmerList.Remove(tmp_Player);
                                    }

                                }
                                //...добаляем игрока в список и задаем случайное время до объявления войны либо есл.
                                else { AI_TooClose_TimmerList.Add(tmp_Player, AI_Randomizer.Next(AI_TooClose_TimmerDefault_Min, AI_TooClose_TimmerDefault_Max)); }
                            }

                            //...если требуется объявить войну...
                            if (Need_War) {

                                //...регистрируем врадждебные действия и зовем на помощь, периодическое выполнение этой операции не позволит игроку отсидеться рядом с враждебным грузовым кораблем дожидаясь окончания войны...
                                RegisterHostileAction(tmp_Faction, CalmdownTime);
                                if (!FleeTimersTriggered) { TriggerFleeTimers(); }

                                Digi.Log.Info(GridName + " - DeclareWarToPlayer: " + tmp_Player.DisplayName + ", Distance: " + CP_Entity_Distance + ", Controller: " + tmp_Player.Controller.ControlledEntity.Entity.EntityId + ", Faction: " + tmp_Faction.Name + "_" + tmp_Faction.FactionId);
                            }
                        }
                        //...иначе считаем что игрок отлетел достаточно далеко и удаляем его из списка.
                        else if (AI_TooClose_TimmerList.ContainsKey(tmp_Player)) { AI_TooClose_TimmerList.Remove(tmp_Player); }

                    }
                    //...иначе, если у игрока нет фракции но он присутствует в списке - удаляем его оттуда ибо игроки без фракций и так враждебны
                    else if (AI_TooClose_TimmerList.ContainsKey(tmp_Player)) { AI_TooClose_TimmerList.Remove(tmp_Player); }

                    if (AISessionCore.AI_Debug && AI_TooClose_TimmerList.ContainsKey(tmp_Player)) {

                        string faction_info = "Faction: None";
                        if (tmp_Faction != null) { faction_info = ", Faction: " + tmp_Faction.Name + "_" + tmp_Faction.FactionId + ", WarTimeLeft: " + AI_TooClose_TimmerList[tmp_Player]; }

                        Digi.Log.Info(GridName + " - Player: " + tmp_Player.DisplayName + ", Distance: " + CP_Entity_Distance + ", Controller: " + tmp_Player.Controller.ControlledEntity.Entity.EntityId + faction_info); 
                    }
                }

                //[Обрабатываем логику перемещения в пространстве]
                //Ищем враждебные объекты в заданном радиусе...
				List<CustomTargetData> RadarData = LookForEnemies(FreighterSetup.FleeTriggerDistance);
                if (AISessionCore.AI_Debug) { Digi.Log.Info(GridName + " - RadarDataCount - " + RadarData.Count + ", Range: " + FreighterSetup.FleeTriggerDistance); }
				
                //...если что-то нашли...
                if (RadarData.Count > 0) {

                    //...получаем ближайший враждебный объект с высшим приоритетом...
                    CustomTargetData ClosestTargetData = RadarData.OrderBy(x => x._Priority).ThenBy(x => x._Distance).FirstOrDefault();
                        
                    //...если это ни ничто...
                    if (!ClosestTargetData._EntityInfo.IsEmpty()) {

                        //...получаем игровой объект на основе id ближайшего враждебного объекта...
                        IMyEntity EnemyEntity = ClosestTargetData._Entity as IMyEntity;

                        //...если удалось получить объект...
                        if (EnemyEntity != null) {

                            //...определяем дистанцию до ближайшего вражеского объекта...
                            double DistanceToEnemy = GridPosition.DistanceTo(ClosestTargetData._EntityInfo.Position);

                            //...вычисляем почку в противоположной стороне...
                            Vector3D RunPoint = GridPosition.InverseVectorTo(ClosestTargetData._EntityInfo.Position, AI_Fly_Distance);

                            //...рандомизируем ограничение скорости дабы выровняться с кораблем было еще сложнее...
                            float FlySpeed = (float)AI_Randomizer.Next((int)(AI_RunSpeed_Min*100), (int)(AI_RunSpeed_Max*100))/100f;
                            
                            //...и летим от него подальше...
                            AI_FlyToPoint(RunPoint, FlySpeed, "RunFrom: " + EnemyEntity.DisplayName + "_" + EnemyEntity.EntityId.ToString());

                            //...зовем на помощь если еще не звали.
                            if (!FleeTimersTriggered) { TriggerFleeTimers(); }

                            if (AISessionCore.AI_Debug) { 
                                Digi.Log.Info(GridName + " - FreighterAI_Mode - Fly away from target: " + EnemyEntity.DisplayName + "_" + ClosestTargetData._EntityInfo.EntityId + ", MaxFlySpeed: " + FlySpeed + ", Distance: " + ClosestTargetData._Distance); 
                            }


                        }
                    }
 
                }
                //...иначе летим себе не спеши...
                else { 

                    //...получаем вектор направления блока удаленного управления, перемножаем на дистанцию получая тем самым удаленную точку
                    Vector3D RetreatPoint = RC.WorldMatrix.Forward * AI_Fly_Distance; 

                    //...в режиме свободного полета выставляем максимум скорости значитлеьно ниже от капа, ибо летим спокойно, не спешим...
                    float FlySpeed = (float)AI_Randomizer.Next((int)(AI_DeepSpaceSpeed_Min*100), (int)(AI_DeepSpaceSpeed_Max*100))/100f;

                    //...пызываем функцию полета к сладосному диспавну
                    AI_FlyToPoint(RetreatPoint, FlySpeed, "DeepSpacePoint");

                    if (AISessionCore.AI_Debug) { Digi.Log.Info(GridName + " - FreighterAI_Mode - Fly to deep space, MaxFlySpeed: " + FlySpeed); }
                
                }

			}
			catch (Exception ex) {
				Digi.Log.Error(ex.Message + Environment.NewLine + ex.StackTrace);
			}
        }
		
        //Функция полета к заданной точке
        public void AI_FlyToPoint(Vector3D TargetPoint, float MaxSpeed, string TargetPointName = "FlyTargetPoint") {

            //Очищаем списой вэйпоинтов автопилота и добавляем нужный
            RC.ClearWaypoints();
		    RC.AddWaypoint(TargetPoint, TargetPointName);

            //Задаем параметры полета и активируем автопилот
		    (RC as MyRemoteControl).ChangeFlightMode(MyRemoteControl.FlightMode.OneWay);
		    (RC as MyRemoteControl).SetAutoPilotSpeedLimit(MaxSpeed);
            (RC as MyRemoteControl).SetCollisionAvoidance(true);
		    RC.SetAutoPilotEnabled(true);
        }

        private void JumpAway()
        {
            List<IMyJumpDrive> JumpDrives = Term.GetBlocksOfType<IMyJumpDrive>(collect: x => x.IsWorking);

            if (JumpDrives.Count > 0) JumpDrives.First().Jump(false);
        }

        private void TriggerFleeTimers()
        {
            if (FleeTimersTriggered) return;

            List<IMyTimerBlock> FleeTimers = new List<IMyTimerBlock>();
            Term.GetBlocksOfType(FleeTimers, x => x.IsFunctional && x.Enabled && (x.CustomName.Contains("Flee") || x.CustomData.Contains("Flee")));
            Grid.DebugWrite("TriggerFleeTimers", $"Flee timers found: {FleeTimers.Count}.{(FleeTimers.Count > 0 ? " Trying to activate..." : "")}");
            foreach (IMyTimerBlock Timer in FleeTimers)
                Timer.Trigger();

            FleeTimersTriggered = true;
        }

        private float DetermineFleeSpeed()
        {
            return Math.Min(FreighterSetup.FleeSpeedCap, FreighterSetup.FleeSpeedRatio * RC.GetSpeedCap());
        }

        protected override bool ParseSetup()
        {
            if (BotBase.ReadBotType(RC) != BotType) return false;

            List<string> CustomData = RC.CustomData.Trim().Replace("\r\n", "\n").Split('\n').ToList();
            foreach (string DataLine in CustomData)
            {
                if (DataLine.Contains("EEM_AI")) continue;
                if (DataLine.Contains("Type")) continue;
                var Data = DataLine.Trim().Split(':');
                Data[1] = Data[1].Trim();
                switch (Data[0].Trim())
                {
                    case "Faction":
                        break;
                    case "FleeOnlyWhenDamaged":
                        if (!bool.TryParse(Data[1], out FreighterSetup.FleeOnlyWhenDamaged))
                        {
                            DebugWrite("ParseSetup", "AI setup error: FleeOnlyWhenDamaged cannot be parsed");
                            return false;
                        }
                        break;
                    case "FleeTriggerDistance":
                        if (!float.TryParse(Data[1], out FreighterSetup.FleeTriggerDistance))
                        {
                            DebugWrite("ParseSetup", "AI setup error: FleeTriggerDistance cannot be parsed");
                            return false;
                        }
                        break;
                    case "FleeSpeedRatio":
                        if (!float.TryParse(Data[1], out FreighterSetup.FleeSpeedRatio))
                        {
                            DebugWrite("ParseSetup", "AI setup error: FleeSpeedRatio cannot be parsed");
                            return false;
                        }
                        break;
                    case "FleeSpeedCap":
                        if (!float.TryParse(Data[1], out FreighterSetup.FleeSpeedCap))
                        {
                            DebugWrite("ParseSetup", "AI setup error: FleeSpeedCap cannot be parsed");
                            return false;
                        }
                        break;
                    default:
                        DebugWrite("ParseSetup", $"AI setup error: Cannot parse '{DataLine}'");
                        return false;
                }
            }

            FreighterSetup.Default();
            return true;
        }

        protected override void DebugWrite(string Source, string Message, string DebugPrefix = "FreighterBot.")
        {
            base.DebugWrite(Source, Message, DebugPrefix);
        }
    }
}