﻿using System;
using System.Text;
using Sandbox.ModAPI;
using VRage.Game.Components;
using VRage.ModAPI;
using Sandbox.Common.ObjectBuilders;
using VRage.ObjectBuilders;
using Sandbox.Definitions;
using VRage.Game;
using Sandbox.Game.EntityComponents;
using VRage.Game.Entity;
using Sandbox.Game;
using VRageMath;
using SpaceEngineers.Game.ModAPI;
using VRage.Game.ModAPI;
using System.Collections.Generic;
using VRage.Collections;
using Sandbox.Game.Entities;
using Sandbox.ModAPI.Interfaces;
using System.Diagnostics;
using System.Linq;
using System.Timers;

namespace Cheetah.AI
{
    [MySessionComponentDescriptor(MyUpdateOrder.BeforeSimulation)]
    public class AISessionCore : MySessionComponentBase
    {
        #region Settings
        // Here are the options which can be freely edited.

        /// <summary>
        /// This toggles the Debug mode. Without Debug, only critical messages are shown in chat.
        /// </summary>
        public static readonly bool Debug = false;

        //Если включен режим дебага то в info.log будут записываться "мысли" AI для последующего анализа
        public static bool AI_Debug = false;

        /// <summary>
        /// This toggles the Debug Anti-Spam mode. With this, a single combination of ship name + message will be displayed only once per session.
        /// </summary>
        public static readonly bool DebugAntiSpam = true;
        #endregion




        public const string WARPIN_EFFECT = "EEMWarpIn";
        Timer DefaultPeaceResetTimmer = new Timer();
        Timer FactionClock = new Timer();
        List<long> NewFactionsIDs = new List<long>();

        /// <summary>
        /// This permits certain operations to throw custom exceptions in order to
        /// provide detailed descriptions of what gone wrong, over call stack.<para />
        /// BEWARE, every exception thrown must be explicitly provided with a catcher, or it will crash the entire game!
        /// </summary>
        public static readonly bool AllowThrowingErrors = true;

        public static bool IsServer
        {
            get
            {
                return !MyAPIGateway.Multiplayer.MultiplayerActive || MyAPIGateway.Multiplayer.IsServer;
            }
        }

        private static Dictionary<long, BotBase.OnDamageTaken> DamageHandlers = new Dictionary<long, BotBase.OnDamageTaken>();
        #region DictionaryAccessors
        static public void AddDamageHandler(long GridID, BotBase.OnDamageTaken Handler)
        {
            DamageHandlers.Add(GridID, Handler);
        }
        static public void AddDamageHandler(IMyCubeGrid Grid, BotBase.OnDamageTaken Handler)
        {
            AddDamageHandler(Grid.GetTopMostParent().EntityId, Handler);
        }
        static public void RemoveDamageHandler(long GridID)
        {
            DamageHandlers.Remove(GridID);
        }
        static public void RemoveDamageHandler(IMyCubeGrid Grid)
        {
            RemoveDamageHandler(Grid.GetTopMostParent().EntityId);
        }
        static public bool HasDamageHandler(long GridID)
        {
            return DamageHandlers.ContainsKey(GridID);
        }
        static public bool HasDamageHandler(IMyCubeGrid Grid)
        {
            return HasDamageHandler(Grid.GetTopMostParent().EntityId);
        }
        #endregion

        bool Inited = false;
        bool FactionsInited = false;

        public override void UpdateBeforeSimulation()
        {
            //if (MyAPIGateway.Multiplayer.MultiplayerActive && !MyAPIGateway.Multiplayer.IsServer) return;
            if (Inited) return;

            InitNPCFactions();
            MyAPIGateway.Session.DamageSystem.RegisterBeforeDamageHandler(0, DamageRefHandler);
            MyAPIGateway.Session.DamageSystem.RegisterDestroyHandler(0, GenericDamageHandler);
            Inited = true;
        }

        public void DamageRefHandler(object DamagedObject, ref MyDamageInformation Damage)
        {
            GenericDamageHandler(DamagedObject, Damage);
        }

        public void GenericDamageHandler(object DamagedObject, MyDamageInformation Damage) {
            try {

                if (DamagedObject == null) { return; }
                if (!(DamagedObject is IMySlimBlock)) { return; }
                IMySlimBlock DamagedBlock = DamagedObject as IMySlimBlock;
                IMyCubeGrid DamagedGrid = DamagedBlock.CubeGrid;
                long GridID = DamagedGrid.GetTopMostParent().EntityId;

                if (DamageHandlers.ContainsKey(GridID)) {
                    string GridName = String.Format("{0}_{1}", DamagedGrid.DisplayName, DamagedGrid.EntityId.ToString());
                    try { DamageHandlers[GridID].Invoke(DamagedBlock, Damage); }
                    //catch (Exception Scrap) { LogError(GridName + " - DamageHandler.Invoke", Scrap); }
                    /*
                    Тихий перехват, я и так знаю что ошибка есть, она не критична и пока я не найду где вылезает гребаный нуль нет смысла выводить ее клиентам
                    Object reference not set to an instance of an object.
                       at Cheetah.AI.BotBase.<Init>b__56_2(IMySlimBlock Block, MyDamageInformation Damage)
                       at Cheetah.AI.AISessionCore.GenericDamageHandler(Object DamagedObject, MyDamageInformation Damage)
                    */
                    catch {} 
                }

            }
            catch (Exception Scrap) { LogError("GenericDamageHandler", Scrap); }
        }

        //Одноклеточный обработчик команд чата.
        private void Utilities_MessageEntered(string messageText, ref bool sendToOthers) {

            //Включем либо отключаем режим дебага для AI и предотвращаем отсылку этого технического сообщения другим игрокам
            if (messageText.ToLower() == "/EEM AI_Debug: Enable".ToLower()) {
                AI_Debug = true;
                Digi.Log.Info("AI_Debug: ENABLED");
                DebugHelper.Print("SessionCore", "EEM AI_Debug: ENABLED, data write into info.log in storage", false);
                sendToOthers = false;
            }
            else if (messageText.ToLower() == "/EEM AI_Debug: Disable".ToLower()) {
                AI_Debug = false;
                Digi.Log.Info("AI_Debug: DISABLED");
                DebugHelper.Print("SessionCore", "EEM AIDebug: DISABLED", false);
                sendToOthers = false;
            }

            
        }
/*
public void SpawnPrefab ( string shipName,  Vector3 pos) {            
            //looks for the definition of the ship
            var prefab =  MyDefinitionManager.Static.GetPrefabDefinition(shipName);
 
            //get the grid containing the ship
            var grid = prefab.CubeGrids[0];

            //give the grid containing the ship a new position
            grid.PositionAndOrientation = new VRage.MyPositionAndOrientation(pos, Vector3.Forward, Vector3.Up);

            //important ! every entity has a unique ID. If you try to add your prefab twice, they would have the same ID so that wont work
            //so you use RemapObjectBuilder to generate new IDs for the ship and other entities onboard.
            MyAPIGateway.Entities.RemapObjectBuilder(grid);
 
            //Create the object and add it to the world.
            var entity = MyAPIGateway.Entities.CreateFromObjectBuilderAndAdd(grid);
}
*/
        private void InitNPCFactions()
        {
            try
            {
                //Для сервера даем возможность использовать команды чата
                if (AISessionCore.IsServer) { MyAPIGateway.Utilities.MessageEntered += Utilities_MessageEntered; }

                Diplomacy.Init();
                var FactionSystem = MyAPIGateway.Session.Factions;

                HashSet<IMyFaction> FactionsToMakePeaceWith = Diplomacy.LawfulFactions.ToHashSet();
                FactionsToMakePeaceWith.UnionWith(FactionSystem.Factions.Values.Where(x => x.IsPlayerFaction()));
                foreach (var NPC_Faction in Diplomacy.LawfulFactions)
                {
                    PeaceTimers.Add(NPC_Faction, new Dictionary<IMyFaction, DateTime>());
                    foreach (var Other_Faction in FactionsToMakePeaceWith)
                    {
                        if (NPC_Faction == Other_Faction) continue;
                        if (NPC_Faction.IsPeacefulTo(Other_Faction)) continue;

                        if (Other_Faction.IsPlayerFaction()) {
                            DateTime PeaceTime = DateTime.Now.Add(Diplomacy.DefaultPeaceTime);

                            var timerdict = PeaceTimers[NPC_Faction];

                            if (!timerdict.ContainsKey(Other_Faction)) { timerdict.Add(Other_Faction, PeaceTime); }
                            else { timerdict[Other_Faction] = PeaceTime; }
                        }
                        else { NPC_Faction.ProposePeace(Other_Faction, Print: true); }
                    }
                }
                
                #region Peace Reset Delay
                //Таймер для сброса дефолтового мира
                DefaultPeaceResetTimmer.Interval = 50;
                DefaultPeaceResetTimmer.AutoReset = false;
                DefaultPeaceResetTimmer.Elapsed += (trash1, trash2) =>
                {

                    foreach (long ID in NewFactionsIDs) {
                        //Поскольку по умолчанию все неписе-фракции автоматом предлагают мир для новой фракции, исправляем это безобразие
                        foreach (IMyFaction LawfulFaction in Diplomacy.LawfulFactions) {
                            MyAPIGateway.Session.Factions.CancelPeaceRequest(LawfulFaction.FactionId, ID);
                            MyAPIGateway.Session.Factions.DeclareWar(LawfulFaction.FactionId, ID);
                        }
                    }
                    NewFactionsIDs.Clear();

                    DefaultPeaceResetTimmer.Stop();

                };
                
                FactionSystem.FactionCreated += (FactionID) =>
                {
                    //Добавляем новую фракцию в список и запускаем таймер для отмены базового авто-мира -_-
                    NewFactionsIDs.Add(FactionID);
                    DefaultPeaceResetTimmer.Start();

                    //Получаем новую фракцию и добавляем ей двухминутный кулдаун на предложение мира, дабы не халявили пересоздавая фракию и тут же заключая мир
                    IMyFaction NewFaction = FactionSystem.TryGetFactionById(FactionID);
                    if (NewFaction != null) {
                        foreach (IMyFaction LawfulFaction in Diplomacy.LawfulFactions) {
                            DateTime PeaceTime = DateTime.Now.Add(TimeSpan.FromMinutes(2));

                            var timerdict = PeaceTimers[LawfulFaction];

                            if (!timerdict.ContainsKey(NewFaction)) { timerdict.Add(NewFaction, PeaceTime); }
                            else { timerdict[NewFaction] = PeaceTime; }
                        }
                    }

                };
                #endregion

                #region Faction Clock
                //Таймер заключения мира после объявления войны
                FactionClock.Interval = 10000;
                FactionClock.AutoReset = true;
                FactionClock.Elapsed += (trash1, trash2) =>
                {
                    if (!FactionsInited)
                    {
                        foreach (var Faction1 in Diplomacy.LawfulFactions)
                        {
                            foreach (var Faction2 in Diplomacy.LawfulFactions)
                            {
                                if (Faction1 == Faction2) continue;
                                if (Faction1.IsPeacefulTo(Faction2)) continue;
                                Faction1.AcceptPeace(Faction2);
                            }
                        }
                        FactionsInited = true;
                    }

                    foreach (var NPCFactionTimer in PeaceTimers)
                    {
                        var NPCFaction = NPCFactionTimer.Key;
                        List<IMyFaction> remove = new List<IMyFaction>();
                        foreach (var Timer in NPCFactionTimer.Value)
                        {
                            var Faction = Timer.Key;
                            if (DateTime.Now >= Timer.Value)
                            {
                                NPCFaction.ProposePeace(Faction, Print: true);
                                remove.Add(Faction);
                            }
                        }

                        foreach (var Faction in remove)
                            NPCFactionTimer.Value.Remove(Faction);
                    }
                };
                FactionClock.Start();
                #endregion
            }
            catch (Exception Scrap)
            {
                LogError("InitNPCFactions", Scrap);
            }
        }

        public static Dictionary<IMyFaction, Dictionary<IMyFaction, DateTime>> PeaceTimers = new Dictionary<IMyFaction, Dictionary<IMyFaction, DateTime>>();
        public static void DeclareWar(IMyFaction OwnFaction, IMyFaction HostileFaction, TimeSpan TruceDelay)
        {
            if (!PeaceTimers.ContainsKey(OwnFaction))
            {
                if (AllowThrowingErrors) throw new Exception($"PeaceTimers dictionary error: can't find {OwnFaction.Tag} key!");
                return;
            }

            DateTime PeaceTime = DateTime.Now.Add(TruceDelay);

            var timerdict = PeaceTimers[OwnFaction];

            if (!timerdict.ContainsKey(HostileFaction))
            {
                timerdict.Add(HostileFaction, PeaceTime);
            }
            else
            {
                timerdict[HostileFaction] = PeaceTime;
            }

            if (!OwnFaction.IsHostileTo(HostileFaction))
            {
                OwnFaction.DeclareWar(HostileFaction, Print: true);
            }

            DebugWrite($"DeclareWarTimer[{OwnFaction.Tag}]", $"Added peace timer. Current time: {DateTime.Now.ToString("HH:mm:ss")} | Calmdown at: {PeaceTime.ToString("HH:mm:ss")} | Calmdown delay: {TruceDelay.ToString()}");
        }

        public static void LogError(string Source, Exception Scrap, string DebugPrefix = "SessionCore.") {
            //Дублируем исключение в лог ибо принтить только в игрвой чат имхо не лучшая идея...
            Digi.Log.Info(String.Format("[Exception] {1}{0}{2}{0}{3}{0}{4}",
                Environment.NewLine,
                DebugPrefix + Source,
                Scrap.Message,
                Scrap.StackTrace,
                (Scrap.InnerException != null ? "[InnerException]" + Scrap.InnerException.Message : "No inner data")
            ));
            DebugHelper.Print("SessionCore", $"Fatal error in '{DebugPrefix + Source}': {Scrap.Message}. {(Scrap.InnerException != null ? Scrap.InnerException.Message : "No additional info was given by the game :(")}");
        }

        public static void DebugWrite(string Source, string Message, bool AntiSpam = true) {

            if (Debug) {
                //Дублируем запись в лог ибо принтить только в игрвой чат имхо не лучшая идея...
                Digi.Log.Info(Source +" - "+ Message);
                DebugHelper.Print($"{Source}", $"{Message}", AntiSpam);
            }
        }
    }
}